(* -*- coq-prog-args: ("-top" "TemplateTestSuite.demo") -*-  *)

Require Import List Arith.
Require Import Template.monad_utils Template.Ast Template.Template.
Import ListNotations MonadNotation.

Local Open Scope string_scope.

(** This is just printing **)
Test Quote (fun x : nat => x).

Test Quote (fun (f : nat -> nat) (x : nat) => f x).

Test Quote (let x := 2 in
            match x with
              | 0 => 0
              | S n => n
            end).

(** Build a definition **)
Definition d : Ast.term.
  let t := constr:(fun x : nat => x) in
  let k x := refine x in
  quote_term t k.
Defined.

(** To quote existing definitions **)
Definition id_nat : nat -> nat := fun x => x.

Quote Definition d'' := Eval compute in id_nat.


(** Fixpoints **)
Fixpoint add (a b : nat) : nat :=
  match a with
    | 0 => b
    | S a => S (add a b)
  end.
Eval vm_compute in add.

Quote Definition add_syntax := Eval compute in add.
Quote Definition add_syntax2 := Eval compute in add_syntax.

(** Reflecting definitions **)
Make Definition zero_from_syntax :=
  (Ast.tConstruct (Ast.mkInd "Coq.Init.Datatypes.nat" 0) 0 []).

Make Definition add_from_syntax :=
  ltac:(let t:= eval compute in add_syntax in exact t).


(** Reflecting  Inductives *)

Definition one_i : one_inductive_entry :=
{|
  mind_entry_typename := "demoBool";
  mind_entry_arity := tSort Universe.type0;
  mind_entry_template := false;
  mind_entry_consnames := ["demoTrue"; "demoFalse"];
  mind_entry_lc := [tRel 1; tRel 1];
|}.

Definition one_i2 : one_inductive_entry :=
{|
  mind_entry_typename := "demoBool2";
  mind_entry_arity := tSort Universe.type0;
  mind_entry_template := false;
  mind_entry_consnames := ["demoTrue2"; "demoFalse2"];
  mind_entry_lc := [tRel 0; tRel 0];
|}.

Definition mut_i : mutual_inductive_entry :=
{|
  mind_entry_record := None;
  mind_entry_finite := Finite;
  mind_entry_params := [];
  mind_entry_inds := [one_i; one_i2];
  mind_entry_universes := Monomorphic_ctx ([], Constraint.empty);
  mind_entry_private := None;
|}.

Make Inductive mut_i.
Print demoBool.

(** Putting the above commands in monadic program *)

Run TemplateProgram (tmBind (tmQuote (3 + 3)) tmPrint).

Run TemplateProgram (tmBind (tmQuoteRec add) tmPrint).

Definition printInductive (name  : ident): TemplateMonad unit :=
  (tmBind (tmQuoteInductive name) tmPrint).

Run TemplateProgram (printInductive "Coq.Init.Datatypes.nat").
Run TemplateProgram (printInductive "nat").

Run TemplateProgram (tmBind (tmQuoteConstant "add" false) tmPrint).

Run TemplateProgram (t <- tmLemma "foo4" nat ;;
                     tmDefinition "foo5" (t + t + 2)).
Next Obligation.
  exact 3.
Defined.
Print foo5.

Run TemplateProgram (t <- tmLemma "foo44" nat ;;
                     qt <- tmQuote t ;;
                     t <- tmEval all t ;;
                     tmPrint qt ;; tmPrint t).
Next Obligation.
  exact (3+2).
Defined.

(** Monadic programs on the syntax ("untyped" compared to Mtac) *)
Definition printConstant (name  : ident): TemplateMonad unit :=
  X <- tmUnquote (tConst name []) ;;
  X' <- tmEval all (projT2 X) ;;
 tmPrint X'.

Fail Run TemplateProgram (printInductive "Coq.Arith.PeanoNat.Nat.add").
Run TemplateProgram (printConstant "Coq.Arith.PeanoNat.Nat.add").

(* Evaluation *)
Run TemplateProgram (tmBind (tmEval all (3 + 3)) tmPrint).
Run TemplateProgram (tmBind (tmEval hnf (3 + 3)) tmPrint).


(* Skipping universes (WIP but supported in syntax) *)
